package com.vincenzojuniorforte.poliutodroid.common;

import android.content.Context;
import android.content.Intent;
import android.graphics.Rect;
import android.os.Bundle;
import android.support.test.InstrumentationRegistry;
import android.support.test.uiautomator.By;
import android.support.test.uiautomator.UiDevice;
import android.support.test.uiautomator.UiObject;
import android.support.test.uiautomator.UiObjectNotFoundException;
import android.support.test.uiautomator.UiSelector;
import android.support.test.uiautomator.Until;
import android.util.Log;
import android.view.KeyEvent;
import android.view.accessibility.AccessibilityEvent;

import com.vincenzojuniorforte.poliutodroid.Config;
import com.vincenzojuniorforte.poliutodroid.FileLog;
import com.vincenzojuniorforte.poliutodroid.core.GuiInteractionController;
import com.vincenzojuniorforte.poliutodroid.model.GuiState;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.Locale;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static com.vincenzojuniorforte.poliutodroid.Config.sOutputDir;
import static com.vincenzojuniorforte.poliutodroid.util.Utility.randomIntBounded;

/**
 * UiAutomator helper
 * <p>
 * GuiHelper offers simplified management features
 * such as watcher registration, dialog handle and human-like input APIs
 */
public class GuiHelper {
    private static final String TAG = Config.TAG;
    private static final String TAG_MAIN = Config.TAG_MAIN;

    // NOTE: Sometimes HOME and BACK key presses will generate no events if already on home page or there is nothing to go back to, Set low timeouts.
    private static final long KEY_PRESS_EVENT_TIMEOUT = 1 * 1000;

    private static final String DUMPSYS_ACTIVITY_ACTIVITIES = "dumpsys activity activities";
    private static final String DUMPSYS_ACTIVITY_TOP = "dumpsys activity top";

    private static Pattern sFocusedActivityPattern =
            Pattern.compile("mFocusedActivity\\: ActivityRecord\\{(.+) u(\\d+) (\\S+) (\\S+)\\}");

    private static Pattern sTopActivityPattern =
            Pattern.compile("ACTIVITY (\\S+) (.+) pid=(\\d+)");

    private static GuiInteractionController sGuiInteractionController = new GuiInteractionController();

    public static int sScreenshotIndex = 0;

    public static String sLastFilename;

    public static GuiWatchers sGuiWatchers = new GuiWatchers();

    /**
     * Register watchers
     */
    public static void registerAnrAndCrashWatchers() {
        sGuiWatchers.registerAnrAndCrashWatchers();
    }

    /**
     * Handle common Android UI (e.g., ANR or crash)
     *
     * @return true if a common Android UI has been handled; false otherwise.
     */
    public static boolean handleAndroidUi() {
        if (sGuiWatchers.handleAnr()) {
            takeScreenshots("[ANR]");
            return true;
        } else if (sGuiWatchers.handleAnr2()) {
            takeScreenshots("[ANR]");
            return true;
        } else if (sGuiWatchers.handleCrash()) {
            takeScreenshots("[CRASH]");
            return true;
        } else if (sGuiWatchers.handleCrash2()) {
            takeScreenshots("[CRASH]");
            return true;
        } else {
            // Something we don't know
        }

        return false;
    }

    /**
     * Check if the last action has been a long click
     *
     * @return true if last action has been a long click; false otherwise.
     */
    // FIXME: doesn't work: UiAutomator doesn't recognize these popup's buttons
    public static boolean longClickIsPerforming() {
        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        UiObject button = null;
        for (String keyword : Config.COMMON_LONG_CLICK_STRINGS) {
            button = device.findObject(new UiSelector().text(keyword).enabled(true));
            if (button != null && button.exists()) {
                return true;
            }
        }

        return false;
    }

    /**
     * Handle common dialog
     *
     * @return true if a common dialog has been handled; false otherwise.
     */
    public static boolean handleCommonDialog() {
        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        UiObject button = null;
        for (String keyword : Config.COMMON_BUTTONS) {
            button = device.findObject(new UiSelector().text(keyword).enabled(true));
            if (button != null && button.exists()) {
                break;
            }
        }
        try {
            // sometimes it takes a while for the OK button to become enabled
            if (button != null && button.exists()) {
                button.waitForExists(5000);
                clickHumanLike(button);
                Log.i("PoLiUToDroidAction", "{Click} " + button.getText() + " Button succeeded");
                return true; // triggered
            }
        } catch (UiObjectNotFoundException e) {
            Log.w(TAG, "UiObject disappear");
        }
        return false; // no trigger
    }

    /**
     * Click at the center of the visible bounds of the UI element represented
     * by this UiObject.
     *
     * @return true if successful else false
     * @throws UiObjectNotFoundException
     */
    public static boolean clickAtTheCenter(UiObject uiObject) throws UiObjectNotFoundException {
        long waitForActionAcknowledgment = 3 * 1000;

        Rect rect = uiObject.getVisibleBounds();
        return new GuiInteractionController().clickAndSync(rect.centerX(), rect.centerY(),
                waitForActionAcknowledgment);
    }

    /**
     * Click at a random point within the visible bounds of the UI element represented
     * by this UiObject.
     *
     * @return true if successful else false
     * @throws UiObjectNotFoundException
     */
    public static boolean clickHumanLike(UiObject uiObject) throws UiObjectNotFoundException {
        Rect rect = uiObject.getVisibleBounds();
        return sGuiInteractionController.clickAndSync(randomIntBounded(rect.left, rect.right), randomIntBounded(rect.bottom, rect.top),
                Config.sWaitForActionAcknowledgment);
    }

    /**
     * Long click the center of the visible bounds of the UI element
     *
     * @return true if operation was successful
     * @throws UiObjectNotFoundException
     * @since API Level 16
     */
    public static boolean longClickAtTheCenter(UiObject uiObject) throws UiObjectNotFoundException {
        Rect rect = uiObject.getVisibleBounds();
        return sGuiInteractionController.longTapAndSync(rect.centerX(), rect.centerY(),
                Config.sWaitForActionAcknowledgment);
    }

    /**
     * Long click at a random point within the visible bounds of the UI element
     *
     * @return true if operation was successful
     * @throws UiObjectNotFoundException
     * @since API Level 16
     */
    public static boolean longClickHumanLike(UiObject uiObject) throws UiObjectNotFoundException {
        Rect rect = uiObject.getVisibleBounds();
        return sGuiInteractionController.longTapAndSync(randomIntBounded(rect.left, rect.right), randomIntBounded(rect.bottom, rect.top),
                Config.sWaitForActionAcknowledgment);
    }

    /**
     * Fill all EditText with randoms pre-defined strings
     */
    public static void inputRandomTextToEditText() {
        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        UiObject edit;

        int i = 0;
        do {
            edit = null;
            edit = device.findObject(new UiSelector().className("android.widget.EditText").instance(i++));
            if (edit != null && edit.exists()) {
                try {
                    Random rand = new Random();
                    if (Config.sFillAllEditText || rand.nextBoolean()) {
                        // NOTE: allow to maintain some edit box empty
                        String text = Config.RANDOM_TEXT[rand.nextInt(Config.RANDOM_TEXT.length - 1)];
                        edit.setText(text);
                    }
                } catch (UiObjectNotFoundException e) {
                    // Don't worry
                }
            }
        } while (edit != null && edit.exists());
    }

    /**
     * Check if the current screen belongs to the AUT
     *
     * @return true if the current screen belongs to the AUT; false otherwise.
     */
    public static boolean isInTargetApp() {
        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        String pkg = device.getCurrentPackageName();
        if (pkg != null && 0 == pkg.compareToIgnoreCase(Config.sTargetPackage)) {
            return true;
        }
        return false;
    }

    /**
     * Check if a screen belongs to an ignored Activity
     *
     * @param screen object to be verified
     * @return true if the screen belongs to an ignored Activity; false otherwise.
     */
    public static boolean isInIgnoredActivity(GuiState screen) {
        return isInIgnoredActivity(screen.getActivityName());
    }

    /**
     * Check if an Activity is ignored
     *
     * @param activityName Activity name to be verified
     * @return true if the Activity name is ignored; false otherwise.
     */
    public static boolean isInIgnoredActivity(String activityName) {
        for (String ignore : Config.IGNORED_ACTIVITY) {
            if (0 == ignore.compareTo(activityName)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Check if a Gui State is equal to the current state
     *
     * @param target Gui State to be verified
     * @return true if the Gui State verified is equal to the current state; false otherwise.
     */
    public static boolean isInTheSameScreen(GuiState target) {
        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
//        UiObject root = device.findObject(new UiSelector().packageName(Config.sTargetPackage));
        UiObject root = device.findObject(new UiSelector().packageName(Config.sTargetPackage).instance(0)); // TODO: to be verified
        if (root == null || !root.exists()) {
            Log.e(TAG, "Fail to get screen root object");
            return false;
        }
        GuiState current = new GuiState();
        return current.equals(target);
    }

    /**
     * Launch the AUT
     *
     * @return
     */
    public static boolean launchTargetApp() {
        return launchApp(Config.sTargetPackage);
    }

    /**
     * Launch the CrawlerApp
     */
    public static void launchCrawlerApp() {
        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        device.pressHome();
        FileLog.i(TAG_MAIN, "{Launch} " + Config.sCrawlerPackage);
        Context context = InstrumentationRegistry.getContext();
        Intent intent = context.getPackageManager().getLaunchIntentForPackage(Config.sCrawlerPackage);
        context.startActivity(intent);
    }

    /**
     * Launch an application installed on the connected device
     *
     * @param targetPackage application package
     * @return true if the application has been launched successfully; false otherwise.
     */
    public static boolean launchApp(String targetPackage) {
        FileLog.i(TAG_MAIN, "{Launch} " + targetPackage);

        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        String launcherPackage = device.getLauncherPackageName();
        if (launcherPackage.compareToIgnoreCase(targetPackage) == 0) {
            launchHome();
            return true;
        }

        Context context = InstrumentationRegistry.getContext();
        // Returns a "good" intent to launch a front-door activity in a package or null if nothing is found.
        final Intent intent = context.getPackageManager().getLaunchIntentForPackage(targetPackage);
        if (intent != null) {
            intent.addFlags(Intent.FLAG_ACTIVITY_CLEAR_TASK); // Make sure each launch is a new task
            context.startActivity(intent);
            device.wait(Until.hasObject(By.pkg(targetPackage).depth(0)), Config.sLaunchTimeout);
        } else {
            String err = String.format("(%s) No launchable Activity.\n", targetPackage);
            Log.e(TAG, err);
            Bundle bundle = new Bundle();
            bundle.putString("ERROR", err);
            InstrumentationRegistry.getInstrumentation().finish(1, bundle); // FIXME: 15/06/2017 Define resultCode
        }
        return true;
    }

    /**
     * Press the Back button
     */
    public static void pressBack() {
        FileLog.i(TAG_MAIN, "{Click} Back");

        sGuiInteractionController.sendKeyAndWaitForEvent(KeyEvent.KEYCODE_BACK, 0, AccessibilityEvent.TYPE_WINDOW_CONTENT_CHANGED,
                KEY_PRESS_EVENT_TIMEOUT);

    }

    /**
     * Press the Home button
     */
    public static void launchHome() {
        FileLog.i(TAG_MAIN, "{Press} Home");

        UiDevice uidevice = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
//        uidevice.pressHome();     // NOTE: Ad Hoc Home pressing
        sGuiInteractionController.sendKeyAndWaitForEvent(KeyEvent.KEYCODE_HOME, 0, AccessibilityEvent.TYPE_WINDOW_CONTENT_CHANGED,
                KEY_PRESS_EVENT_TIMEOUT);

        String launcherPackage = uidevice.getLauncherPackageName();
        uidevice.wait(Until.hasObject(By.pkg(launcherPackage).depth(0)), Config.sLaunchTimeout);
    }

    /**
     * Transform a string to a valid file name replacing all non-allowed characters
     *
     * @param input string to be validated
     * @return valid file name
     */
    public static String toValidFileName(String input) {
        if (input == null)
            return "";
        return input.replaceAll("[:\\\\/*\"?|<>']", "_");
    }

    /**
     * Take screenshot of the current device screen
     *
     * @param message
     */
    // TODO: Take screenshot for both portrait and landscape
    public static void takeScreenshots(String message) {
        //Log.v(TAG, new Exception().getStackTrace()[0].getMethodName() + "()");

        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        device.waitForIdle(Config.sWaitIdleTimeout);

//        String activity = device.getCurrentActivityName(); // FIXME: deprecated
        String activity = getTopActivityName();
        if (activity == null)
            activity = "No Activity activityName";
        if (activity.length() > 30) {
            activity = activity.substring(0, 29);
        }

        // Dump window hierarchy for debug, remove it for better performance
        try {
            device.dumpWindowHierarchy(new File(String.format(Locale.getDefault(), "%s/%d - %s.xml", sOutputDir, sScreenshotIndex, activity)));
        } catch (IOException e) {
            e.printStackTrace();
        }

        if (message.length() > 50) {
            message = message.substring(0, 49);
        }

        sLastFilename = "";
        if (message.length() > 0) {
            sLastFilename = String.format(Locale.getDefault(), "(%d) %s %s.png",
                    sScreenshotIndex, toValidFileName(activity), toValidFileName(message));
        } else {
            sLastFilename = String.format(Locale.getDefault(), "(%d) %s.png",
                    sScreenshotIndex, toValidFileName(activity));
        }
        device.takeScreenshot(new File(sOutputDir + "/" + sLastFilename));
        sScreenshotIndex++;
        FileLog.i(TAG_MAIN, "{Screenshot} " + sLastFilename);
    }


    /**
     * Get the focused Activity name
     *
     * @return Focused Activity name
     * @Deprecated Use getTopActivityName because faster
     */
    @Deprecated
    public static String getFocusedActivityName(UiDevice device, String pkg) {
        String sysDump = null;

        if (pkg == null)
            return null;

        try {
            sysDump = device.executeShellCommand(DUMPSYS_ACTIVITY_ACTIVITIES);
            if (sysDump == null || !sysDump.contains(pkg)) {
                return null;
            }
        } catch (IOException e) {
            e.printStackTrace();

            return null;
        }

        LinkedList<String> mSysDump = new LinkedList<>();
        Collections.addAll(mSysDump, sysDump.split("\\n"));

        while (!mSysDump.isEmpty()) {

            final String line = mSysDump.pop().trim();
            Matcher matcher = sFocusedActivityPattern.matcher(line);
            if (matcher.matches()) {
                int endIndex = matcher.group(3).lastIndexOf(".");
                if (endIndex != -1)
                    return matcher.group(3).substring(endIndex + 1);    // Remove "."
            }
        }

        return null;
    }

    /**
     * Get the focused Activity Name
     *
     * @return Focused Activity name
     */
    public static String getTopActivityName() {
        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        device.waitForIdle(Config.sWaitIdleTimeout);

        String pkg = device.getCurrentPackageName();
        String sysDump = null;

        if (pkg == null)
            return null;

        try {
            sysDump = device.executeShellCommand(DUMPSYS_ACTIVITY_TOP);
            if (sysDump == null || !sysDump.contains(pkg)) {

                return null;
            }
        } catch (IOException e) {
            e.printStackTrace();

            return null;
        }

        LinkedList<String> mSysDump = new LinkedList<>();
        Collections.addAll(mSysDump, sysDump.split("\\n"));

        while (!mSysDump.isEmpty()) {

            final String line = mSysDump.pop().trim();
            Matcher matcher = sTopActivityPattern.matcher(line);
            if (matcher.matches()) {
                int endIndex = matcher.group(1).lastIndexOf(".");
                if (endIndex != -1)
                    return matcher.group(1).substring(endIndex + 1);    // Remove "."
            }
        }

        return null;
    }

    /**
     * Get the full classname of the focused Activity
     *
     * @return Full classname of the focused Activity
     */
    public static String getTopActivityClassName() {
        UiDevice device = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        device.waitForIdle(Config.sWaitIdleTimeout);

        String pkg = device.getCurrentPackageName();
        String sysDump = null;

        if (pkg == null)
            return null;

        try {
            sysDump = device.executeShellCommand(DUMPSYS_ACTIVITY_TOP);
            if (sysDump == null || !sysDump.contains(pkg)) {

                return null;
            }
        } catch (IOException e) {
            e.printStackTrace();

            return null;
        }

        LinkedList<String> mSysDump = new LinkedList<>();
        Collections.addAll(mSysDump, sysDump.split("\\n"));

        while (!mSysDump.isEmpty()) {

            final String line = mSysDump.pop().trim();
            Matcher matcher = sTopActivityPattern.matcher(line);
            if (matcher.matches()) {
                return matcher.group(1).replace("/", "");
            }
        }

        return null;
    }
}
