package com.vincenzojuniorforte.poliutodroid.core;

import android.app.Instrumentation;
import android.app.UiAutomation;
import android.os.SystemClock;
import android.support.test.InstrumentationRegistry;
import android.support.test.uiautomator.Configurator;
import android.support.test.uiautomator.UiDevice;
import android.util.Log;
import android.view.InputDevice;
import android.view.InputEvent;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.ViewConfiguration;
import android.view.accessibility.AccessibilityEvent;

import com.vincenzojuniorforte.poliutodroid.Config;

import java.util.Locale;
import java.util.concurrent.TimeoutException;

import static com.vincenzojuniorforte.poliutodroid.test.BuildConfig.DEBUG;

/**
 * GuiInteractionController class
 * <p>
 * It is responsible for injecting user events such as touch events (includes swipes)
 * and text key events into the system. To do so, all it needs to know about are coordinates
 * (random location inside the widget's surface) of the touch events and text for the text input events.
 * The GuiInteractionController performs no synchronization. It will fire touch and text input events
 * as fast as it receives them. All the events injected contain real (i.e., non-dummy) information.
 */
public class GuiInteractionController {
    private static final long REGULAR_CLICK_LENGTH = 100;
    private static final String LOG_TAG = GuiInteractionController.class.getSimpleName();

    private UiDevice mDevice;
    private Instrumentation mInstrumentation;
    private long mDownTime;

    private static int sDeviceId = Config.sDeviceId == -1 ? 0 : Config.sDeviceId;

    /**
     * GuiInteractionController constructor
     */
    public GuiInteractionController() {
        mInstrumentation = InstrumentationRegistry.getInstrumentation();
        mInstrumentation.waitForIdleSync();
        mDevice = UiDevice.getInstance(mInstrumentation);
        mDevice.waitForIdle(Config.sWaitIdleTimeout);

        if (sDeviceId == 0) {
            for (int id : InputDevice.getDeviceIds()) {
                if (id > 0) {
                    InputDevice dev = InputDevice.getDevice(id);
                    if (dev != null) {
                        int sources = dev.getSources();
                        // Verify that the device has a touchscreen
                        if ((sources & InputDevice.SOURCE_TOUCHSCREEN) == InputDevice.SOURCE_TOUCHSCREEN) {
                            Config.sDeviceId = sDeviceId = id;
                            break;
                        }
                    }
                }
            }
        }
    }

    /**
     * Send keys and blocks until the first specified accessibility event.
     * <p>
     * Most key presses will cause some UI change to occur. If the device is busy, this will
     * block until the device begins to process the key press at which point the call returns
     * and normal wait for idle processing may begin. If no events are detected for the
     * timeout period specified, the call will return anyway with false.
     *
     * @param keyCode   code related to the key pressed
     * @param metaState state of the meta key
     * @param eventType the event type
     * @param timeout   waiting for event
     * @return true if events is received, otherwise false.
     */
    public boolean sendKeyAndWaitForEvent(final int keyCode, final int metaState,
                                          final int eventType, long timeout) {
        Runnable command = new Runnable() {
            @Override
            public void run() {
                final long eventTime = SystemClock.uptimeMillis();
                /*
                long downTime,
                long eventTime,
                int action,
                int code,
                int repeat,
                int metaState,
                int deviceId,
                int scancode,
                int flags,
                int source
                 */
                // NOTE: Scan code constants are raw device-specific codes obtained from the OS and so are not generally meaningful to applications unless interpreted using the KeyCharacterMap.
//                int scanCode = Utility.randomInt(0, 283);

                KeyEvent downEvent = new KeyEvent(eventTime, eventTime, KeyEvent.ACTION_DOWN,
                        keyCode, 0, metaState, sDeviceId, keyCode/*scanCode*//*0*/, 0,
                        InputDevice.SOURCE_KEYBOARD);
                if (injectEventSync(downEvent)) {
                    KeyEvent upEvent = new KeyEvent(eventTime, eventTime, KeyEvent.ACTION_UP,
                            keyCode, 0, metaState, sDeviceId, keyCode/*scanCode*//*0*/, 0,
                            InputDevice.SOURCE_KEYBOARD);
                    injectEventSync(upEvent);
                }
            }
        };

        return runAndWaitForEvents(command, new GuiInteractionController.WaitForAnyEventPredicate(eventType), timeout)
                != null;
    }

    /**
     * Click at coordinates and blocks until either accessibility event TYPE_WINDOW_CONTENT_CHANGED
     * or TYPE_VIEW_SELECTED are received.
     *
     * @param x       ascended location
     * @param y       ordered location
     * @param timeout waiting for event
     * @return true if events are received, else false if timeout.
     */
    public boolean clickAndSync(final int x, final int y, long timeout) {

        String logString = String.format(Locale.getDefault(), "clickAndSync(%d, %d)", x, y);
        Log.d(LOG_TAG, logString);

        return runAndWaitForEvents(clickRunnable(x, y), new WaitForAnyEventPredicate(
                AccessibilityEvent.TYPE_WINDOW_CONTENT_CHANGED |
                        AccessibilityEvent.TYPE_VIEW_SELECTED), timeout) != null;
    }

    /**
     * Long tap at coordinates and blocks until either accessibility event
     * TYPE_WINDOW_CONTENT_CHANGED or TYPE_VIEW_SELECTED are received.
     *
     * @param x       ascended location
     * @param y       ordered location
     * @param timeout waiting for event
     * @return true if events are received, else false if timeout.
     */
    public boolean longTapAndSync(final int x, final int y, long timeout) {

        String logString = String.format(Locale.getDefault(), "clickAndSync(%d, %d)", x, y);
        Log.d(LOG_TAG, logString);

        return runAndWaitForEvents(longTapRunnable(x, y), new WaitForAnyEventPredicate(
                AccessibilityEvent.TYPE_WINDOW_CONTENT_CHANGED |
                        AccessibilityEvent.TYPE_VIEW_SELECTED), timeout) != null;
    }

    private AccessibilityEvent executeCommandAndWaitForAccessibilityEvent(Runnable command,
                                                                          UiAutomation.AccessibilityEventFilter filter, long timeoutMillis) throws TimeoutException {
        return mInstrumentation.getUiAutomation().executeAndWaitForEvent(command,
                filter, timeoutMillis);
    }

    /**
     * Helper used by methods to perform actions and wait for any accessibility events and return
     * predicated on predefined filter.
     *
     * @param command runnable command
     * @param filter  listener for filtering accessibility events
     * @param timeout waiting for event
     * @return accessibility event object
     */
    private AccessibilityEvent runAndWaitForEvents(Runnable command,
                                                   UiAutomation.AccessibilityEventFilter filter, long timeout) {

        try {
            return executeCommandAndWaitForAccessibilityEvent(command, filter,
                    timeout);
        } catch (TimeoutException e) {
            Log.w(LOG_TAG, "runAndwaitForEvents timed out waiting for events");
            return null;
        } catch (Exception e) {
            Log.e(LOG_TAG, "exception from executeCommandAndWaitForAccessibilityEvent", e);
            return null;
        }
    }

    /**
     * Returns a Runnable for use in {@link #runAndWaitForEvents(Runnable, UiAutomation.AccessibilityEventFilter, long) to
     * perform a click.
     *
     * @param x coordinate
     * @param y coordinate
     * @return Runnable object
     */
    private Runnable clickRunnable(final int x, final int y) {
        return new Runnable() {
            @Override
            public void run() {
                if (touchDown(x, y)) {
                    SystemClock.sleep(REGULAR_CLICK_LENGTH);
                    touchUp(x, y);
                }
            }
        };
    }

    /**
     * Returns a Runnable for use in {@link #runAndWaitForEvents(Runnable, UiAutomation.AccessibilityEventFilter, long) to
     * perform a long tap.
     *
     * @param x coordinate
     * @param y coordinate
     * @return Runnable object
     */
    private Runnable longTapRunnable(final int x, final int y) {
        return new Runnable() {
            @Override
            public void run() {
                if (touchDown(x, y)) {
                    SystemClock.sleep(ViewConfiguration.getLongPressTimeout());
                    touchUp(x, y);
                }
            }
        };
    }

    private boolean touchDown(int x, int y) {
        if (DEBUG) {
            Log.d(LOG_TAG, "touchDown (" + x + ", " + y + ")");
        }

        mDownTime = SystemClock.uptimeMillis();
        MotionEvent event = getMotionEvent(mDownTime, mDownTime, MotionEvent.ACTION_DOWN, x, y);

        return injectEventSync(event);
    }

    private boolean touchUp(int x, int y) {
        if (DEBUG) {
            Log.d(LOG_TAG, "touchUp (" + x + ", " + y + ")");
        }

        final long eventTime = SystemClock.uptimeMillis();
        MotionEvent event = getMotionEvent(mDownTime, eventTime, MotionEvent.ACTION_UP, x, y);
        mDownTime = 0;

        return injectEventSync(event);
    }

    /**
     * Helper function to obtain a MotionEvent.
     */
    private MotionEvent getMotionEvent(long downTime, long eventTime, int action,
                                       float x, float y) {

        MotionEvent.PointerProperties properties = new MotionEvent.PointerProperties();
        properties.id = 0;
        properties.toolType = Configurator.getInstance().getToolType();

        if (properties.toolType == MotionEvent.TOOL_TYPE_UNKNOWN)
            properties.toolType = MotionEvent.TOOL_TYPE_FINGER;

        MotionEvent.PointerCoords coords = new MotionEvent.PointerCoords();
        coords.pressure = 1;
        coords.size = 1;
        coords.x = x;
        coords.y = y;

        /*
        long downTime,
        long eventTime,
        int action,
        int pointerCount,
        PointerProperties[] pointerProperties,
        PointerCoords[] pointerCoords,
        int metaState,
        int buttonState,
        float xPrecision,
        float yPrecision,
        int deviceId,
        int edgeFlags,
        int source,
        int flags
         */

        return MotionEvent.obtain(downTime, eventTime, action, 1,
                new MotionEvent.PointerProperties[]{properties}, new MotionEvent.PointerCoords[]{coords},
                0, 0, 1.0f, 1.0f, sDeviceId, 0, InputDevice.SOURCE_TOUCHSCREEN, 0);
    }

    /**
     * Perform an action
     *
     * @param event action to be performed
     * @return true if the action has been performed successfully; false otherwise.
     */
    private boolean injectEventSync(InputEvent event) {
        return mInstrumentation.getUiAutomation().injectInputEvent(event, true);
    }

    private class WaitForAnyEventPredicate implements UiAutomation.AccessibilityEventFilter {
        int mMask;

        WaitForAnyEventPredicate(int mask) {
            mMask = mask;
        }

        @Override
        public boolean accept(AccessibilityEvent t) {
            // check current event in the list
            if ((t.getEventType() & mMask) != 0) {
                return true;
            }

            // no match yet
            return false;
        }
    }
}
