package com.vincenzojuniorforte.poliutodroid.model;


import android.app.Instrumentation;
import android.support.annotation.NonNull;
import android.support.test.uiautomator.UiDevice;
import android.support.test.uiautomator.UiObject;
import android.support.test.uiautomator.UiObjectNotFoundException;
import android.util.Log;

import com.vincenzojuniorforte.poliutodroid.Config;
import com.vincenzojuniorforte.poliutodroid.common.GuiHelper;
import com.vincenzojuniorforte.poliutodroid.model.structure.GenericTree;
import com.vincenzojuniorforte.poliutodroid.core.GuiHierarchyDumper;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import static android.support.test.InstrumentationRegistry.getInstrumentation;
import static com.vincenzojuniorforte.poliutodroid.Config.sWaitIdleTimeout;
import static org.hamcrest.CoreMatchers.notNullValue;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;

/**
 * GUI State class
 * <p>
 * It represents a unique screen (Gui state) that is worth testing. The screen of Android Activity
 * can change dynamically (e.g. using Fragment), thus one Activity may produce more than one GuiState.
 * A GuiState object contains all details about the screen represented (e.g., depth, signature,
 * package and activity to which the screen belongs, etc.)
 */
public class GuiState implements Comparable<GuiState> {

    //region Static Variables
    private static final String TAG = Config.TAG;
    private static final String TAG_MAIN = Config.TAG_MAIN;
    private static final String TAG_DEBUG = Config.TAG_DEBUG;
    //endregion

    //region Variables
    // Main UiAutomator objects
    private UiDevice device;
    private UiObject rootObject; // First UiObject in the screen    // Maybe useless

    // The first Gui event that led to the current Gui state
    private GuiEvent guiEvent;  //NOTE: the first Gui event is the only relevant one, because the subsequent events coming from Gui state different from the first will find the current status as "Finished" (as DFS strategy)

    // Gui state properties
    private String pkg;
    private String activityName;
    private List<List<Integer>> layoutCIS;  // Accumulated indices of layout (un-executable) widgets from root
    private List<GuiWidget> executableWidgets;   // Executable widgets in the screen that we interested in test (e.g. Clickable)

    // NOTE: useful to achieve a specific Gui state by performing all the Gui events
    private List<GuiEvent> path;  // Accumulated GuiEvent from "main" Gui state

    private int id = -1; // State number in the GUI Graph
    private int depth = -1; // Depth in the GUI Graph (useful for non-AUT apps)

    private String signature;    // Use to identify itself between other screens

    // Auxiliary testing variables
    private boolean mFinished = false;    // True if no longer subject to testing

    public int loop = 0; // Avoid infinite loop
    //endregion

    //region Constructors

    /**
     * GuiState empty constructor
     */
    public GuiState() {
        this.guiEvent = new GuiEvent(null, null, null);

        executableWidgets = new ArrayList<>();
        layoutCIS = new ArrayList<>();
        signature = "";

        init(guiEvent);
    }

    /**
     * GuiState constructor used after an event execution
     *
     * @param guiEvent event previously executed
     * @param id       GuiState numerical incremental identifier
     */
    public GuiState(GuiEvent guiEvent, int id) {
        guiEvent.setArrivalGuiState(this);
        this.guiEvent = guiEvent;
        this.id = id;

        executableWidgets = new ArrayList<>();
        layoutCIS = new ArrayList<>();
        signature = "";

        init(guiEvent);
    }
    //endregion

    //region Getters and Setters

    //region Depth

    /**
     * Get the depth from the initial GuiState if the GuiState belongs to the AUT.
     * Get the relative depth from the first foreign GuiState (depth = 1) of the GuiState doesn't belong to the AUT.
     *
     * @return the depth from the initial GuiState if the GuiState belongs to the AUT,
     * or the relative depth from the first foreign GuiState (depth = 1) of the GuiState doesn't belong to the AUT.
     */
    public int getDepth() {
        return depth;
    }

    /**
     * Set the depth of the GuiState
     *
     * @param depth the depth of the GuiState
     */
    public void setDepth(int depth) {
        this.depth = depth;
    }
    //endregion

    //region Signature

    /**
     * Get the signature of the GuiState.
     *
     * @return the signature of the GuiState.
     */
    public String getSignature() {
        return signature;
    }
    //endregion

    //region Path

    /**
     * Get the path (list of GuiEvents) from the initial GuiState
     *
     * @return the path (list of GuiEvents) from the initial GuiState
     */
    public List<GuiEvent> getPath() {
        return path;
    }

    /**
     * Set the path (list of GuiEvents) adding a new GuiEvent
     *
     * @param path     list of GuiEvents
     * @param guiEvent GuiEvent to be add to the path
     */
    private void setPath(List<GuiEvent> path, GuiEvent guiEvent) {
        this.path = path;
        this.path.add(guiEvent);
    }
    //endregion

    //region LayoutCIS

    /**
     * Get the layout CISs
     *
     * @return the layout CISs
     */
    public List<List<Integer>> getLayoutCIS() {
        return layoutCIS;
    }
    //endregion

    //region ExecutableCIS

    /**
     * Get the executable CISs
     *
     * @return the executable CISs
     */
    public List<List<Integer>> getExecutableCIS() {
        List<List<Integer>> executableCIS = new ArrayList<>();

        for (GuiWidget guiWidget : executableWidgets) {
            executableCIS.add(guiWidget.getCIS());
        }
        return executableCIS;
    }
    //endregion

    //region ExecutableWidgets

    /**
     * Get the executable widgets which belong to the GuiState
     *
     * @return the executable widgets which belong to the GuiState
     */
    public List<GuiWidget> getExecutableWidgets() {
        return executableWidgets;
    }
    //endregion

    //region Pkg

    /**
     * Get the package of which the GuiState belongs to
     *
     * @return the package of which the GuiState belongs to
     */
    public String getPkg() {
        return pkg;
    }
    //endregion

    //region ActivityName

    /**
     * Get the Activity name of which the GuiState belongs to
     *
     * @return the Activity name of which the GuiState belongs to
     */
    public String getActivityName() {
        return activityName;
    }
    //endregion

    //region Id

    /**
     * Get the GuiState id
     *
     * @return the GuiState id
     */
    public int getId() {
        return id;
    }
    //endregion

    //region ParentGuiState

    /**
     * Get the parent GuiState: first GuiState that led to discover the current GuiState
     *
     * @return the parent GuiState
     */
    public GuiState getParentGuiState() {
        return guiEvent.getParentGuiState();
    }
    //endregion

    //region TriggeredWidget

    /**
     * Get the widget that firstly led to discover the current GuiState
     *
     * @return the widget that firstly led to discover the current GuiState
     */
    public GuiWidget getTriggeredWidget() {
        return guiEvent.getTriggeredWidget();
    }
    //endregion

    //region TriggeredAction

    /**
     * Get the action that firsly led to discover the current GuiState
     *
     * @return the action that firstly led to discover the current GuiState
     */
    public GuiWidget.Action getTriggeredAction() {
        return guiEvent.getAction();
    }
    //endregion

    //region Finished

    /**
     * Set the GuiState as finished (no more testable) or unfinished (still testable)
     *
     * @param finished true value for setting the GuiState as finished;
     *                 false value for setting the GuiState as unfinished.
     */
    public void setFinished(boolean finished) {
        mFinished = finished;
    }

    /**
     * Verify if the GuiState is finished
     *
     * @return true if the GuiState is finished; false otherwise.
     */
    public boolean isFinished() {
        if (mFinished) {
            setFinished(true);
            return true;
        }

        for (GuiWidget child : executableWidgets) {
            if (!child.isFinished()) {
                mFinished = false;
                return false;
            }
        }
        mFinished = true;
        return true;
    }
    //endregion
    //endregion

    //region Overrides and Implementations
    @Override
    public String toString() {
        // FIXME: Better to use StringBuilder
        String str = "activityName:" + activityName +
                ", id:" + id +
                ", depth:" + depth +
                ", finished:" + mFinished +
                ", signature:" + signature +
                ", widgets:" + executableWidgets.size();
        for (int i = 0; i < executableWidgets.size(); i++) {
            GuiWidget widget = executableWidgets.get(i);
            str += " " + i + ":" + widget.isFinished();
        }
        return str;
    }

    @Override
    public int compareTo(@NonNull GuiState other) {
        return signature.compareTo(other.getSignature());
//        return Integer.valueOf(id).compareTo(other.getId());
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        GuiState guiState = (GuiState) o;

        return signature.equals(guiState.getSignature());

    }

    @Override
    public int hashCode() {
        return signature.hashCode();
    }
    //endregion

    //region Private Methods

    /**
     * Initialize the GuiState
     *
     * @param guiEvent event object that led to the current GuiState
     */
    private void init(GuiEvent guiEvent) {
        // Get Instrumentation object
        Instrumentation instrumentation = getInstrumentation();

        // Get UiDevice instance
        device = UiDevice.getInstance(instrumentation);

        GenericTree<GuiWidget> widgetTree = null;
        int attempts = 0;
        do {
            device.waitForIdle(sWaitIdleTimeout);

            // Retrieve the package of the foreground Activity
            pkg = device.getCurrentPackageName();

            // Retrieve the foreground Activity name
//        activityName = device.getCurrentActivityName(); // NOTE: deprecated
            activityName = GuiHelper.getTopActivityName();
            if (activityName == null) {
                activityName = "";
            }

            // Instantiate a GuiHierarchyDumper
            GuiHierarchyDumper guiHierarchyDumper = new GuiHierarchyDumper(instrumentation);


            try {
                // Build the widget tree making up the current Gui state
                widgetTree = guiHierarchyDumper.dump(device, pkg);
            } catch (IOException e) {
                e.printStackTrace();
            }

        } while (widgetTree == null && ++attempts < 5);  // NOTE: it copes with transient

        // Assert that widget tree has been built well
        assertThat(widgetTree, notNullValue());
        assertThat(widgetTree.getRoot(), notNullValue());
        assertThat(widgetTree.getRoot().getData(), notNullValue());

        UiObject root =
                widgetTree.getRoot()    // Set tree's root (Node object)
                        .getData()      // Retrieve related GuiWidget object
                        .getUiObject(); // Retrieve related getUiObject

        assertThat(root, notNullValue());
        rootObject = root;

        // Collect widgets
        if(Config.sOnlyFirstChildren) {
            collectWidgets(widgetTree.getRoot());
        } else {
            collectAllWidgets(widgetTree.getRoot());
        }

        GuiState parentGuiState = guiEvent.getParentGuiState();
        if (parentGuiState == null) {
            depth = 0;
            path = new ArrayList<>();
        } else {
            depth = parentGuiState.getDepth() + 1;
            setPath(parentGuiState.getPath(), guiEvent);
        }

        // Build screen signature
        assertTrue(parseSignature());

        // Nothing testable
        if (executableWidgets.size() == 0)
            mFinished = true;

        // Debug
        if (Config.sDebug) {
            Log.d(TAG_DEBUG, signature);
        }
    }

    /**
     * Recursively differentiate and collect widget.
     * Container widget will collect at most one direct-child (with all its sub-elements).
     * Nested containers are well-managed.
     *
     * @param node
     */
    private void collectWidgets(GenericTree<GuiWidget>.Node node) {
        GuiWidget guiWidget = node.getData();

        if (guiWidget == null)
            return;

        // Store information of the current widget
        if (guiWidget.isExecutable())
            executableWidgets.add(guiWidget);
        else
            layoutCIS.add(guiWidget.getCIS());

        if (guiWidget.isContainer()) {
            if (node.getNumberOfChildren() > 0) {
                collectWidgets(node.getChildAt(0));   // Recursion only on first child
                node.getData().setFirstOfList(true);
                return; // At this point first element (with all its sub-elements) has been add; Return on parent node.
            }
        }

        for (GenericTree<GuiWidget>.Node child : node.getChildren()) {
            // Recursion on all children
            collectWidgets(child);
        }
    }

    /**
     * Recursively differentiate and collect widget.
     * Container widget will collect all direct-children (with all its sub-elements).
     * Nested containers are well-managed.
     *
     * @param node
     */
    private void collectAllWidgets(GenericTree<GuiWidget>.Node node) {
        GuiWidget guiWidget = node.getData();

        if (guiWidget == null)
            return;

        // Store information of the current widget
        if (guiWidget.isExecutable())
            executableWidgets.add(guiWidget);
        else
            layoutCIS.add(guiWidget.getCIS());

        if (guiWidget.isContainer()) {
            for(int i = 0; i < node.getNumberOfChildren(); i++) {
                // The properties of the container will be inherited by its children
                GenericTree<GuiWidget>.Node childNode = node.getChildAt(i);
//                GuiWidget childGuiWidget = childNode.getData();
//                childGuiWidget.setInheritedExecutableProperties(node.getData());

                collectAllWidgets(childNode);   // Recursion on children
                if(i == 0) {
                    node.getData().setFirstOfList(true);
                }
            }

            return; // At this point all children (with all its sub-elements) has been add; Return on parent node.
        }

        for (GenericTree<GuiWidget>.Node child : node.getChildren()) {
            // Recursion on all children
            collectAllWidgets(child);
        }
    }

    /**
     * Build the GuiState signature based on the comparison criteria and the comparison level choose
     *
     * @return true if the signature has been successfully built; false otherwise.
     */
    private boolean parseSignature() {
        if ((rootObject == null) || !rootObject.exists())
            return false;

        StringBuilder stringBuilder = new StringBuilder("");
        switch (Config.sComparisonLevel) {
            case 5:
                // Contents comparison level

                for (GuiWidget guiWidget : executableWidgets) {
                    String text = "";
                    String desc = "";

                    try {
                        text = guiWidget.getUiObject().getText();
                        if (text.length() > 25) {
                            text = text.substring(0, 21);
                        }
                    } catch (UiObjectNotFoundException e) {
                        // Don't worry, it is ok without Text
                    }
                    try {
                        desc = guiWidget.getUiObject().getContentDescription();
                    } catch (UiObjectNotFoundException e) {
                        // Never mind, it is ok without ContentDescription
                    }

                    stringBuilder.append(text).append(desc);
                }

            case 4:
                // Executable widget comparison level

                String trigSignatures = "";
                // Concat executable property values
                for (GuiWidget guiWidget : executableWidgets) {
                    trigSignatures += guiWidget.getExecPropertyValue();
                }
                stringBuilder.insert(0, ":" + trigSignatures);
            case 3:
                // Layout widget comparison level

                stringBuilder.insert(0, ":" + layoutCIS.toString());
            case 2:
                // Activity activityName comparison level

                stringBuilder.insert(0, ":" + activityName);
            case 1:
                // Package comparison level

                stringBuilder.insert(0, pkg);
        }

        signature = stringBuilder.toString();

        return true;
    }
    //endregion

    //region Public Methods

    /**
     * Reset the states of all executable widgets belonging to the Gui state
     */
    public void reset() {
        mFinished = false;

        for (GuiWidget child : executableWidgets) {
            child.resetExecutableProperties();
        }

        loop = 0;
    }

    /**
     * Restore the state of a specific executable widget belonging to the Gui state
     *
     * @param executableWidget target widget
     * @param triggeredAction  action to be restored on widget
     */
    public void restoreActionWidget(GuiWidget executableWidget, GuiWidget.Action triggeredAction) {
        for (GuiWidget guiWidget : getExecutableWidgets()) {
            if (guiWidget.equals(executableWidget)) {
                if (guiWidget.restoreAction(triggeredAction)) {
                    setFinished(false);
                }
            }
        }
    }
    //endregion
}
