package com.vincenzojuniorforte.poliutodroid.model;

import android.support.annotation.NonNull;
import android.support.test.uiautomator.UiObject;
import android.support.test.uiautomator.UiObjectNotFoundException;
import android.support.test.uiautomator.UiScrollable;
import android.support.test.uiautomator.UiSelector;
import android.util.Log;
import android.view.accessibility.AccessibilityNodeInfo;

import com.vincenzojuniorforte.poliutodroid.Config;

import java.util.ArrayList;
import java.util.List;

import static com.vincenzojuniorforte.poliutodroid.common.GuiHelper.clickHumanLike;
import static com.vincenzojuniorforte.poliutodroid.common.GuiHelper.longClickHumanLike;

/**
 * GUI Widget class
 * <p>
 * It represents an android view in the GuiState that we are interested in test (e.g. Button)
 * - Clickable;
 * - Long Clickable;
 * - Checkable;
 * - Scrollable;
 * <p>
 * A GuiWidget contains details about the widget represented (e.g., characteristics and properties)
 */
public class GuiWidget implements Comparable<GuiWidget> {

    //region Variables & Enums
    // Main UiObject
    private UiObject uiObject;

    // Cumulative (from the root widget) index sequences
    private List<Integer> CIS;

    // Widget properties
    private boolean clickable, longClickable, checkable, scrollable;
    private boolean scrolledDown = false;
    private int executablePropertyValue = 0;

    // Container widget properties
    private boolean isContainer = false;
    private boolean firstOfList = false;

    // Not Accessibility Friendly
    private boolean NAF = false;

    // Auxiliary testing variables
    private int mFullTested = 0;
    private boolean mFinished; // True if no longer subject to testing


    // Executable property values

    /**
     * Enumerator: action values
     */
    public enum Action {
        CLICK(1), LONG_CLICK(2), CHECK(4), SCROLL(8);

        private int value;

        Action(int value) {
            this.value = value;
        }
    }
    //endregion

    //region Constructors

    /**
     * GuiWidget constructor
     *
     * @param node     AccessibilityNodeInfo characterizing the GuiWidget
     * @param uiObject UiObject characterizing the GuiWidget
     */
    public GuiWidget(@NonNull AccessibilityNodeInfo node, @NonNull UiObject uiObject) {
//        this.node = node;
        this.uiObject = uiObject;
        this.CIS = new ArrayList<>();

        // Set widget properties & mFinished
        setClickable(node.isClickable());
//        setScrollable(node.isScrollable());   // TODO: to be managed
        setCheckable(node.isCheckable());
        setLongClickable(node.isLongClickable());

        // Set container widget properties
        for (String containerClass : Config.CONTAINER_CLASS) {
            if (node.getClassName().toString().compareTo(containerClass) == 0) {
                isContainer = true;
                break;
            }
        }

        executablePropertyValue = mFullTested;  // Fixed once at the beginning: sum of executable properties
    }
    //endregion

    //region Getters and Setters
    //region UiObject

    /**
     * Get the UiObject characterizing the GuiWidget
     *
     * @return the UiObject characterizing the GuiWidget
     */
    public UiObject getUiObject() {
        return uiObject;
    }
    //endregion

    //region Container

    /**
     * Verify if the GuiWidget is a container object
     *
     * @return true if the GuiWidget is a container object; false otherwise.
     */
    public boolean isContainer() {
        return isContainer;
    }

    /**
     * Verify if the container GuiWidget has at least one child object
     *
     * @return
     */
    boolean haveFirstOfList() {
        return isContainer && firstOfList;
    }

    /**
     * Set if the container GuiWidget has at least one child object
     *
     * @param firstOfList value which determines whether the container GuiWidget has at least one child object
     */
    void setFirstOfList(boolean firstOfList) {
        if (isContainer())
            this.firstOfList = firstOfList;
    }
    //endregion

    //region Clickable

    /**
     * Check if the GuiWidget is clickable
     *
     * @return true if the GuiWidget is clickable; false otherwise.
     */
    public boolean isClickable() {
        return clickable;
    }

    /**
     * Set if the GuiWidget is clickable
     *
     * @param clickable value which determines whether the GuiWidget is clickable
     */
    private void setClickable(boolean clickable) {
        this.clickable = clickable;
        if (clickable)
            mFullTested += Action.CLICK.value;
    }
    //endregion

    //region LongClickable

    /**
     * Check if the GuiWidget is long clickable
     *
     * @return true if the GuiWidget is long clickable; false otherwise.
     */
    public boolean isLongClickable() {
        return longClickable;
    }

    /**
     * Set if the GuiWidget is long clickable
     *
     * @param longClickable value which determines whether the GuiWidget is long clickable
     */
    private void setLongClickable(boolean longClickable) {
        this.longClickable = longClickable;
        if (longClickable)
            mFullTested += Action.LONG_CLICK.value;
    }
    //endregion

    //region Checkable

    /**
     * Check if the GuiWidget is checkable
     *
     * @return true if the GuiWidget is checkable; false otherwise.
     */
    public boolean isCheckable() {
        return checkable;
    }

    /**
     * Set if the GuiWidget is checkable
     *
     * @param checkable value which determines whether the GuiWidget is checkable
     */
    private void setCheckable(boolean checkable) {
        this.checkable = checkable;
        if (checkable)
            mFullTested += Action.CHECK.value;
    }
    //endregion

    //region Scrollable

    /**
     * Check if the GuiWidget is scrollable
     *
     * @return true if the GuiWidget is scrollable; false otherwise.
     */
    public boolean isScrollable() {
        return scrollable;
    }

    /**
     * Set if the GuiWidget is scrollable
     *
     * @param scrollable value which determines whether the GuiWidget is scrollable
     */
    private void setScrollable(boolean scrollable) {
        this.scrollable = scrollable;
        if (scrollable)
            mFullTested += Action.SCROLL.value;
    }
    //endregion

    //region Finished

    /**
     * Set the GuiWidget as finished (no more testable through some actions) or unfinished (still testable through some actions)
     *
     * @param finished true value for setting the GuiWidget as finished;
     *                 false value for setting the GuiWidget as unfinished.
     */
    public void setFinished(boolean finished) {
        mFinished = finished;
    }

    /**
     * Verify if the GuiWidget is finished
     *
     * @return true if the GuiWidget is finished; false otherwise.
     */
    public boolean isFinished() {
        return mFinished;
    }
    //endregion

    //region Executable

    /**
     * Inherit executable properties from its parent (container)
     *
     * @param container GuiWidget container (parent) from which inheriting the executable properties
     */
    void setInheritedExecutableProperties(GuiWidget container) {
        setClickable(container.isClickable());
        setLongClickable(container.isLongClickable());
    }

    /**
     * Get the executable property value (sum of the values of the actions triggerable on the UI component) of the GuiWidget
     *
     * @return
     */
    public int getExecPropertyValue() {
        return executablePropertyValue;
    }

    /**
     * Verify if the GuiWidget is executable (i.e., some action can be performed on it)
     *
     * @return true if the the GuiWidget is executable; false otherwise.
     */
    public boolean isExecutable() {
        return executablePropertyValue != 0;
    }
    //endregion

    //region FullTested

    /**
     * Verify if the GuiWidget is full tested (i.e., all available actions triggerable on the UI component have been performed)
     *
     * @return true if the GuiWidget is full tested; false otherwise
     */
    public boolean isFullTested() {
        return mFullTested == 0;
    }
    //endregion

    //region CIS

    /**
     * Get the CIS characterizing the GuiWidget
     *
     * @return the CIS characterizing the GuiWidget
     */
    public List<Integer> getCIS() {
        return CIS;
    }

    /**
     * Set the CIS characterizing the GuiWidget
     *
     * @param CIS the CIS characterizing the GuiWidget
     */
    public void setCIS(List<Integer> CIS) {
        this.CIS = new ArrayList<>(CIS);
    }

    /**
     * Add an index at the CIS characterizing the GuiWidget
     *
     * @param index index to be added to the CIS characterizing the GuiWidget
     */
    public void setCIS(int index) {
        this.CIS.add(index);
    }

    /**
     * Set the CIS characterizing the GuiWidget and add the an index to it
     *
     * @param CIS   the CIS characterizing the GuiWidget
     * @param index index to be added to the CIS characterizing the GuiWidget
     */
    public void setCIS(List<Integer> CIS, int index) {
        this.CIS = new ArrayList<>(CIS);
        this.CIS.add(index);
    }
    //endregion

    //region NAF

    /**
     * Verify if the GuiWidget is Not Accessibility Friendly
     *
     * @return true if the GuiWidget is Not Accessibility Friendly; false otherwise.
     */
    public boolean isNAF() {
        return NAF;
    }

    /**
     * Set if the GuiWidget is Not Accessibility Friendly
     *
     * @param NAF true value for setting the GuiWidget as NAF
     */
    public void setNAF(boolean NAF) {
        this.NAF = NAF;
    }
    //endregion
    //endregion

    //region Overrides and Implementations
    @Override
    public int compareTo(@NonNull GuiWidget other) {
        return Integer.valueOf(executablePropertyValue).compareTo(other.getExecPropertyValue()) + CIS.toString().compareTo(other.getCIS().toString());
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        GuiWidget guiWidget = (GuiWidget) o;

        return executablePropertyValue == guiWidget.getExecPropertyValue() && CIS.toString().equals(guiWidget.getCIS().toString());
    }

    @Override
    public int hashCode() {
        int result = CIS.toString().hashCode();
        result = 31 * result + executablePropertyValue;
        return result;
    }
    //endregion

    //region Private Methods

    /**
     * Check if an action can be performed on the GuiWidget
     *
     * @param action action to be verified
     * @return true if the action can be performed on the GuiWidget; false otherwise
     */
    private boolean checkExecutableProperty(Action action) {
        return (executablePropertyValue & action.value) != 0;
    }

    /**
     * Verify if the GuiWidget is still testable by click
     *
     * @return true if the GuiWidget is still testable by click; false otherwise
     */
    private boolean isTestableByClick() {
        return (mFullTested & Action.CLICK.value) != 0;
    }

    /**
     * Verify if the GuiWidget is still testable by long click
     *
     * @return true if the GuiWidget is still testable by long click; false otherwise
     */
    private boolean isTestableByLongClick() {
        return (mFullTested & Action.LONG_CLICK.value) != 0;
    }

    /**
     * Verify if the GuiWidget is still testable by check
     *
     * @return true if the GuiWidget is still testable by check; false otherwise
     */
    private boolean isTestableByCheck() {
        return (mFullTested & Action.CHECK.value) != 0;
    }

    /**
     * Verify if the GuiWidget is still testable by scroll
     *
     * @return true if the GuiWidget is still testable by scroll; false otherwise
     */
    private boolean isTestableByScroll() {
        return (mFullTested & Action.SCROLL.value) != 0;
    }
    //endregion

    //region Package-private Methods

    /**
     * Reset the state of the widget
     */
    void resetExecutableProperties() {
        mFinished = false;
        mFullTested = executablePropertyValue;
    }

    /**
     * Restore the state of a the widget
     *
     * @param triggeredAction action to be restored on widget
     * @return true if the recovery is successful, otherwise false
     */
    boolean restoreAction(Action triggeredAction) {
        if (checkExecutableProperty(triggeredAction)) {
            // Restore feasibility test
            mFullTested |= triggeredAction.value;
            setFinished(false);
            return true;
        }

        return false;
    }
    //endregion

    //region Public Methods

    /**
     * Test the widget with the next action available
     *
     * @return the action performed or null if no action has been performed
     * @throws UiObjectNotFoundException
     */
    public Action test() throws UiObjectNotFoundException {
        Action lastActionPerformed = null;

        if (isFullTested()) {
            setFinished(true);
            return null;
        }

        if (isTestableByClick()) {
            // Sometimes it takes a while for a widget to become enabled
            if (uiObject != null && uiObject.exists()) {
                Log.i("PoLiUToDroidAction", "{Click} " + uiObject.getText() + " Widget succeeded");
                uiObject.waitForExists(5000);
                clickHumanLike(uiObject);

                mFullTested -= Action.CLICK.value;

                lastActionPerformed = Action.CLICK;
            }
        } else if (isTestableByLongClick()) {
            // Sometimes it takes a while for a widget to become enabled
            if (uiObject != null && uiObject.exists()) {
                Log.i("PoLiUToDroidAction", "{LongClick} " + uiObject.getText() + " Widget succeeded");
                uiObject.waitForExists(5000);
//                uiObject.longClick();
                longClickHumanLike(uiObject);

                mFullTested -= Action.LONG_CLICK.value;

                lastActionPerformed = Action.LONG_CLICK;
            }
        } else if (isTestableByCheck()) {
            // Sometimes it takes a while for a widget to become enabled
            if (uiObject != null && uiObject.exists()) {
                Log.i("PoLiUToDroidAction", "{Check} " + uiObject.getText() + " Widget succeeded");
                uiObject.waitForExists(5000);
                clickHumanLike(uiObject);

                mFullTested -= Action.CHECK.value;

                lastActionPerformed = Action.CHECK;
            }
        } else if (isTestableByScroll()) {
            // TODO: to be managed (also swipe???)
            // Sometimes it takes a while for a widget to become enabled
            if (uiObject != null && uiObject.exists()) {
                Log.i("PoLiUToDroidAction", "{Scroll} " + uiObject.getText() + " Widget succeeded");
                uiObject.waitForExists(5000);

                UiScrollable uiScrollable = new UiScrollable(new UiSelector().className(uiObject.getClassName()));
//                UiScrollable uiScrollable  = new UiScrollable(uiObject.getSelector());
                if (scrolledDown) {
                    uiScrollable.scrollToBeginning(100);
//                    uiObject.swipeUp(1);
                } else {
                    uiScrollable.scrollToEnd(100);
//                    uiObject.swipeDown(1);
                }

                mFullTested -= Action.SCROLL.value;
                scrolledDown = !scrolledDown;

                lastActionPerformed = Action.SCROLL;
            }
        }

        if (isFullTested()) {
            setFinished(true);
        }

        return lastActionPerformed;
    }
    //endregion
}
