package com.vincenzojuniorforte.poliutodroid.model.structure;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class GenericTree<T extends Comparable<T>> {

    private Node root;

    public enum TraversalOrderEnum {
        PRE_ORDER,
        POST_ORDER
    }

    public GenericTree() {
        super();
    }

    public GenericTree(Node root) {
        setRoot(root);
    }

    public Node getRoot() {
        return this.root;
    }

    public void setRoot(Node root) {
        this.root = root;
    }

    public int getNumberOfNodes() {
        int numberOfNodes = 0;

        if (root != null) {
            numberOfNodes = auxiliaryGetNumberOfNodes(root) + 1; //1 for the root!
        }

        return numberOfNodes;
    }

    private int auxiliaryGetNumberOfNodes(Node node) {
        int numberOfNodes = node.getNumberOfChildren();

        for (Node child : node.getChildren()) {
            numberOfNodes += auxiliaryGetNumberOfNodes(child);
        }

        return numberOfNodes;
    }

    public boolean exists(T dataToFind) {
        return (find(dataToFind) != null);
    }

    public Node find(T dataToFind) {
        Node returnNode = null;

        if (root != null) {
            returnNode = auxiliaryFind(root, dataToFind);
        }

        return returnNode;
    }

    private Node auxiliaryFind(Node currentNode, T dataToFind) {
        Node returnNode = null;
        int i = 0;

        if (currentNode.getData() != null && currentNode.getData().compareTo(dataToFind) == 0) {
            returnNode = currentNode;
        } else if (currentNode.hasChildren()) {
            i = 0;
            while (returnNode == null && i < currentNode.getNumberOfChildren()) {
                returnNode = auxiliaryFind(currentNode.getChildAt(i), dataToFind);
                i++;
            }
        }

        return returnNode;
    }

    public boolean isEmpty() {
        return (root == null);
    }

    public List<Node> build(TraversalOrderEnum traversalOrder) {
        List<Node> returnList = null;

        if (root != null) {
            returnList = build(root, traversalOrder);
        }

        return returnList;
    }

    public List<Node> build(Node node, TraversalOrderEnum traversalOrder) {
        List<Node> traversalResult = new ArrayList<Node>();

        if (traversalOrder == TraversalOrderEnum.PRE_ORDER) {
            buildPreOrder(node, traversalResult);
        } else if (traversalOrder == TraversalOrderEnum.POST_ORDER) {
            buildPostOrder(node, traversalResult);
        }

        return traversalResult;
    }

    private void buildPreOrder(Node node, List<Node> traversalResult) {
        traversalResult.add(node);

        for (Node child : node.getChildren()) {
            buildPreOrder(child, traversalResult);
        }
    }

    private void buildPostOrder(Node node, List<Node> traversalResult) {
        for (Node child : node.getChildren()) {
            buildPostOrder(child, traversalResult);
        }

        traversalResult.add(node);
    }

    public Map<Node, Integer> buildWithDepth(TraversalOrderEnum traversalOrder) {
        Map<Node, Integer> returnMap = null;

        if (root != null) {
            returnMap = buildWithDepth(root, traversalOrder);
        }

        return returnMap;
    }

    public Map<Node, Integer> buildWithDepth(Node node, TraversalOrderEnum traversalOrder) {
        Map<Node, Integer> traversalResult = new LinkedHashMap<Node, Integer>();

        if (traversalOrder == TraversalOrderEnum.PRE_ORDER) {
            buildPreOrderWithDepth(node, traversalResult, 0);
        } else if (traversalOrder == TraversalOrderEnum.POST_ORDER) {
            buildPostOrderWithDepth(node, traversalResult, 0);
        }

        return traversalResult;
    }

    private void buildPreOrderWithDepth(Node node, Map<Node, Integer> traversalResult, int depth) {
        traversalResult.put(node, depth);

        for (Node child : node.getChildren()) {
            buildPreOrderWithDepth(child, traversalResult, depth + 1);
        }
    }

    private void buildPostOrderWithDepth(Node node, Map<Node, Integer> traversalResult, int depth) {
        for (Node child : node.getChildren()) {
            buildPostOrderWithDepth(child, traversalResult, depth + 1);
        }

        traversalResult.put(node, depth);
    }

    public String toString() {
        /*
        We're going to assume a pre-order traversal by default
         */

        String stringRepresentation = "";

        if (root != null) {
            stringRepresentation = build(TraversalOrderEnum.PRE_ORDER).toString();

        }

        return stringRepresentation;
    }

    public String toStringWithDepth() {
        /*
        We're going to assume a pre-order traversal by default
         */

        String stringRepresentation = "";

        if (root != null) {
            stringRepresentation = buildWithDepth(TraversalOrderEnum.PRE_ORDER).toString();
        }

        return stringRepresentation;
    }

    ///////////////////////
    // Generic Tree Node //
    ///////////////////////

    public class Node {

        private T data;
        private List<Node> children;
        private Node parent;

        public Node() {
            super();
            parent = null;
            children = new ArrayList<>();
        }

        public Node(T data) {
            this();
            setData(data);
        }

        public Node getParent() {
            return this.parent;
        }

        public List<Node> getChildren() {
            return this.children;
        }

        public int getNumberOfChildren() {
            return getChildren().size();
        }

        public int getNumberOfRecursiveChildren() {
            int numberOfNodes = getNumberOfChildren();

            for (Node child : getChildren()) {
                numberOfNodes += child.getNumberOfRecursiveChildren();
            }

            return numberOfNodes;
        }

        public boolean hasChildren() {
            return (getNumberOfChildren() > 0);
        }

        public void setChildren(List<Node> children) {
            for (Node child : children) {
                child.parent = this;
            }

            this.children = children;
        }

        public void addChild(Node child) {
            child.parent = this;
            children.add(child);
        }

        public void addChildAt(int index, Node child) throws IndexOutOfBoundsException {
            child.parent = this;
            children.add(index, child);
        }

        public void removeChildren() {
            this.children = new ArrayList<Node>();
        }

        public void removeChildAt(int index) throws IndexOutOfBoundsException {
            children.remove(index);
        }

        public Node getChildAt(int index) throws IndexOutOfBoundsException {
            return children.get(index);
        }

        public T getData() {
            return this.data;
        }

        public void setData(T data) {
            this.data = data;
        }

        public String toString() {
            return getData().toString();
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            Node other = (Node) obj;
            if (data == null) {
                if (other.data != null) {
                    return false;
                }
            } else if (!data.equals(other.data)) {
                return false;
            }
            return true;
        }

        /* (non-Javadoc)
        * @see java.lang.Object#hashCode()
        */
        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + ((data == null) ? 0 : data.hashCode());
            return result;
        }

        public String toStringVerbose() {
            String stringRepresentation = getData().toString() + ":[";

            for (Node node : getChildren()) {
                stringRepresentation += node.getData().toString() + ", ";
            }

            //Pattern.DOTALL causes ^ and $ to match. Otherwise it won't. It's retarded.
            Pattern pattern = Pattern.compile(", $", Pattern.DOTALL);
            Matcher matcher = pattern.matcher(stringRepresentation);

            stringRepresentation = matcher.replaceFirst("");
            stringRepresentation += "]";

            return stringRepresentation;
        }
    }
}
