/*
 * Depth First + Local Optimization Crawler
 *
 * Android UI flow is a graph with cycles, not a tree, we never know where will we go before clicking a button.
 * Same as our life.
 */

package com.vincenzojuniorforte.poliutodroid.strategy;

import android.graphics.Rect;
import android.support.test.InstrumentationRegistry;
import android.support.test.uiautomator.UiDevice;
import android.support.test.uiautomator.UiObjectNotFoundException;
import android.util.Log;

import com.vincenzojuniorforte.poliutodroid.Config;
import com.vincenzojuniorforte.poliutodroid.FileLog;
import com.vincenzojuniorforte.poliutodroid.common.GuiHelper;
import com.vincenzojuniorforte.poliutodroid.model.GuiEvent;
import com.vincenzojuniorforte.poliutodroid.model.GuiState;
import com.vincenzojuniorforte.poliutodroid.model.GuiWidget;
import com.vincenzojuniorforte.poliutodroid.model.structure.GenericDirectedGraph;
import com.vincenzojuniorforte.poliutodroid.performance.PerformanceMonitor;
import com.vincenzojuniorforte.poliutodroid.util.Utility;

import java.util.Date;
import java.util.Locale;

import static com.vincenzojuniorforte.poliutodroid.Config.sMaxSleepBetweenActions;
import static com.vincenzojuniorforte.poliutodroid.Config.sMinSleepBetweenActions;
import static com.vincenzojuniorforte.poliutodroid.common.GuiHelper.pressBack;

/**
 * PoLiUToDroid test using Android UiAutomator 2.0 and UiAutomation.
 * <p>
 * DepthFirstCrawler implements the E-LTS learning algorithm by following a depth-first strategy.
 * It maintains the GUI model learned during the exploration.
 */
public class DepthFirstCrawler extends Crawler {
    private static final String TAG = Config.TAG;
    private static final String TAG_MAIN = Config.TAG_MAIN;
    private static final String TAG_DEBUG = Config.TAG_DEBUG;

    private static int sDepth = 0;  // root screen depth = 0
    private static int sSteps = 0;
    private static int sDepthPeak = 0;

    private static Date sStartTime;
    private static boolean sRestarted = false;

    private static GenericDirectedGraph<GuiState> sGuiGraph = new GenericDirectedGraph<>();

    private static GuiState sRootScreen = null;
    private static GuiState sLastScreen = null;
    private static GuiWidget sLastActionWidget = null;
    private static GuiWidget.Action sLastAction = null;

    private static String sLastActionMessage = "";

    private static boolean sFinished = false;

    private UiDevice mDevice;

    @Override
    public void run() {
        //region Initialization
        sDepth = 0;
        sSteps = 0;
        sDepthPeak = 0;
        sStartTime = new Date();
        sRootScreen = null;
        sLastScreen = null;
        sLastActionWidget = null;
        sLastAction = null;
        sLastActionMessage = "";
        sFinished = false;

        mDevice = UiDevice.getInstance(InstrumentationRegistry.getInstrumentation());
        //endregion

        // Start from main activity
        if (!GuiHelper.launchTargetApp())
            return;

        Config.sDisplayResolution = mDevice.getDisplayHeight() * mDevice.getDisplayWidth();

        while (!sFinished) {
            sSteps++;

            mDevice.waitForIdle();

            // Build the event (GuiEvent) that led to the current Gui state
            GuiEvent guiEvent = new GuiEvent(sLastScreen, sLastActionWidget, sLastAction);

            // Build current Gui state
            GuiState currentState = new GuiState(guiEvent, sGuiGraph.size() + 1);

            // In other package
            boolean inOtherPackage = false;
            if (currentState.getPkg().compareTo(Config.sTargetPackage) != 0) {
                if (currentState.getPkg().compareTo(mDevice.getLauncherPackageName()) == 0) {
                    // Do not consider Launcher
                    GuiHelper.launchTargetApp();
                    sRestarted = true;
                    sDepth = 0;
                    continue;
                }
                inOtherPackage = true;
                FileLog.i(TAG_MAIN, "{Inspect} screen, in other package: " + currentState.getPkg());
                handleOtherPackage(currentState);
            }

            boolean newScreen = true;
            GuiState storedScreen = sGuiGraph.retrieve(currentState);
            if (storedScreen != null) {
                // Old state
                newScreen = false;
                currentState = storedScreen;

                if (!inOtherPackage) {
                    // AUT
                    if (sRestarted) {
                        sRestarted = false;
                        // Updated stored depth: root Gui state will be overwritten
                        // For example, a first-run-wizard that are no longer visualized.
                        currentState.setDepth(0);
                    } else {
                        sDepth = currentState.getDepth();
                    }
                } else {
                    // In other package
                    sDepth += currentState.getDepth();
                }
            }

            if (sDepth == 0) {  // NOTE: depth of Gui state belonging to other package is always greater than 0
                if (sRootScreen != null && !sRootScreen.equals(currentState)) {
                    Log.i(TAG, "Root screen changed, may be due to app's coachmarks");
                }
                sRootScreen = currentState;
            }

            // New screen?
            if (newScreen) {
                handleNewScreen(currentState, inOtherPackage);
            } else {
                handleOldScreen(currentState, inOtherPackage);
            }

            // Record CPU and Memory info
            PerformanceMonitor.record(currentState.getActivityName());

            // Input some random text to EditText if any
            // TODO: give possibility to insert real user inputs (tester's settings)
            // NOTE: to avoid lost of focus in the case of text selection is performing
//            if (!GuiHelper.longClickIsPerforming()) {
            if (sLastAction != GuiWidget.Action.LONG_CLICK) {
                if (Config.sRandomText) {
                    GuiHelper.inputRandomTextToEditText();
                }
            }

            // Handle next unfinished widget
            handleNextWidget(currentState);

            // Check finish
            if (currentState.isFinished()) {
                Log.d(TAG, "Screen[" + currentState.getId() + "] finished");

                // TODO: sGraph.getRoots().contains(currentState)
                if (currentState.equals(sRootScreen)) {
                    // Root screen is finished
                    if (isNewOrUnfinishedScreen() /*|| !isAllScreensFinished()*/) {
                        sRootScreen.reset();
                    } else {
                        FileLog.i(TAG_MAIN, "{Stop} root screen finished, id:" + sRootScreen.getId());
                        sFinished = true;
                    }
                } else {
                    if (GuiHelper.isInTheSameScreen(currentState)) {
                        // NOTE: if the last action do not change we try to come back to the previous Gui state via Back key
                        pressBack();
                        mDevice.waitForIdle(Config.sWaitIdleTimeout);
                    } else {
                        if (!isNewOrUnfinishedScreen()) {
                            // After action: old and finished Gui state
                            GuiHelper.launchTargetApp();
                            sRestarted = true;
                            sDepth = 0;
                        }
                    }
                }
            } else {
                // If are have unfinished widgets, mark all ascendant screens unfinished, so we have chance to go back.
                // NOTE: possible choices for parent: 1) The one at a lower depth 2) Or the one with the largest id (last visited)
                // NOTE: sLastScreen could be equal to the current screen
                GuiState screen = currentState;
                do {
                    // NOTE: we choice to consider the first parent for each screen in order to ensure a proper DFS strategy
                    if (screen.getParentGuiState() != null && screen.getTriggeredWidget() != null) {
                        screen.getParentGuiState().restoreActionWidget(screen.getTriggeredWidget(), screen.getTriggeredAction());
                    }
                    screen = screen.getParentGuiState();
                } while (screen != null);
            }

            // Max run time
            if ((new Date().getTime() - sStartTime.getTime()) / 1000 > Config.sMaxRuntime) {
                FileLog.i(TAG_MAIN, "{Stop} reached max run-time second: " + Config.sMaxRuntime);
                sFinished = true;
            }

            //  Max screen files
            if (GuiHelper.sScreenshotIndex >= Config.sMaxScreenshot - 1) {
                FileLog.i(TAG_MAIN, "{Stop} reached max screenshot files.");
                sFinished = true;
            }

            // Max test steps
            if (sSteps >= Config.sMaxSteps) {
                FileLog.i(TAG_MAIN, "{Stop} reached max screenshot files.");
                sFinished = true;
            }

            // Avoid infinite loop
            if (++currentState.loop > Config.sMaxScreenLoop) {
                Log.i(TAG, "Reached max screen loop, set screen finished");
                currentState.setFinished(true);
            }

            // Debug
            if (Config.sDebug) {
                logAllScreenInfo();
            }
        }

        // Done
        FileLog.i(TAG_MAIN, "Total executed steps:" + sSteps +
                ", peak depth:" + sDepthPeak +
                ", detected screens:" + sGuiGraph.size() +
                ", screenshot:" + GuiHelper.sScreenshotIndex);

        String log = String.format(Locale.getDefault(), "CPU average:%.1f%%, CPU peak:%.1f%%, " +
                        "Memory average (KB):%d, Memory peak (KB):%d",
                PerformanceMonitor.getAverageCpu(), PerformanceMonitor.cpuPeak,
                PerformanceMonitor.getAverageMemory(), PerformanceMonitor.memPeak);
        FileLog.i(TAG_MAIN, log);
    }

    private void handleOtherPackage(GuiState currentScreen) {
        if (isNewScreen(currentScreen)) {
            GuiHelper.takeScreenshots("(" + currentScreen.getPkg() + ")");

            if (sLastScreen != null) {  // NOTE: should be always not null because test start from the AUT
                if (sLastScreen.getPkg().compareTo(Config.sTargetPackage) == 0) {
                    // Reset the depth outside the target package domain
                    currentScreen.setDepth(1); // NOTE: depth of Gui state belonging to other package is always greater than 0
                } else {
                    // NOTE: depth is well-managed into GuiState class
                    // Update the depth outside the target package domain
//                    currentScreen.setDepth(sLastScreen.getDepth() + 1);
                }
            }

            // Check the depth outside the target package domain
            if (currentScreen.getDepth() > Config.sMaxForeignDepth) {
                // Remove executable widget to be sure will be not executed in future
                currentScreen.getExecutableWidgets().clear();
                // Set Gui state as finished (no longer testable)
                currentScreen.setFinished(true);
            }
        }
    }

    private void handleNewScreen(GuiState currentScreen, boolean inOtherPackage) {
        sLastActionMessage = "";
        sLastActionWidget = null;
        sLastAction = null;

        if (!inOtherPackage) {
            FileLog.i(TAG_MAIN, "{Inspect} NEW screen, " + currentScreen.toString());
            GuiHelper.takeScreenshots("");

            currentScreen.setDepth(++sDepth);
            if (sDepth > sDepthPeak)
                sDepthPeak = sDepth;
        }

        boolean stop = false;
        if (GuiHelper.isInIgnoredActivity(currentScreen.getActivityName())) {
            FileLog.i(TAG_MAIN, "{Inspect} screen, in ignored list: " + currentScreen.getActivityName());
            stop = true;
        }
        if (sDepth >= Config.sMaxDepth) {
            Log.i(TAG, "Has reached the MaxDepth: " + Config.sMaxDepth);
            stop = true;
        }

        if (stop) {
            // Remove executable widget to be sure will be not executed in future
            currentScreen.getExecutableWidgets().clear();
            // Set Gui state as finished (no longer testable)
            currentScreen.setFinished(true);
            pressBack();    // NOTE: Not sure that we can always go back to previous page by Back key
            mDevice.waitForIdle(Config.sWaitIdleTimeout);
        }

        // Add new Gui state
        sGuiGraph.add(sLastScreen, currentScreen);

        if (inOtherPackage) {
            if (GuiHelper.handleAndroidUi()) {
                FileLog.i(TAG_MAIN, "Handle Android UI succeeded");
            } else if (GuiHelper.handleCommonDialog()) {
                FileLog.i(TAG_MAIN, "Handle Common UI succeeded");
            }
        }
    }

    private void handleOldScreen(GuiState currentScreen, boolean inOtherPackage) {
        FileLog.i(TAG_MAIN, "{Inspect} OLD screen, " + currentScreen.toString());
        if (Config.sCaptureSteps) {
            // FIXME: probably, to be removed!
            GuiHelper.takeScreenshots(sLastActionMessage);
            sLastActionMessage = "";
            sLastActionWidget = null;
            sLastAction = null;
        }

        // Update Gui state: add eventually a new edge
        sGuiGraph.update(sLastScreen, currentScreen);

        if (inOtherPackage) {
            if (GuiHelper.handleAndroidUi()) {
                FileLog.i(TAG_MAIN, "Handle Android UI succeeded");
            } else if (GuiHelper.handleCommonDialog()) {
                FileLog.i(TAG_MAIN, "Handle Common UI succeeded");
            }
        }
    }

    private void handleNextWidget(GuiState currentScreen) {
        GuiWidget widget = getNextWidget(currentScreen);
        if (widget == null) {
            return;
        }

        String fullClassName = "";
        String text = "";
        String desc = "";

        try {
            fullClassName = widget.getUiObject().getClassName();
            text = widget.getUiObject().getText();
            if (text.length() > 25) {
                text = text.substring(0, 21) + "...";
            }
        } catch (UiObjectNotFoundException e) {
            // Don't worry, it is ok without Text
        }
        try {
            desc = widget.getUiObject().getContentDescription();
        } catch (UiObjectNotFoundException e) {
            // Never mind, it is ok without ContentDescription
        }
        try {
            Rect bounds = widget.getUiObject().getBounds();
            String className = "";
            for (String tmp : fullClassName.split("\\.")) {
                className = tmp;
            }
            if (text.length() > 0)
                sLastActionMessage = String.format("{Click} %s %s %s", text, className, bounds.toShortString());
            else if (desc.length() > 0)
                sLastActionMessage = String.format("{Click} %s %s %s", desc, className, bounds.toShortString());
            else
                sLastActionMessage = String.format("{Click} %s %s", className, bounds.toShortString());

            FileLog.i(TAG_MAIN, sLastActionMessage);
            sLastScreen = currentScreen;
            sLastActionWidget = widget;

            // Sleep for a while (human-like behaviour)
            Utility.randomDelay(sMinSleepBetweenActions, sMaxSleepBetweenActions);

            // Test widget
            sLastAction = widget.test();
            mDevice.waitForIdle(Config.sWaitIdleTimeout);

        } catch (UiObjectNotFoundException e) {
            Log.e(TAG, "UiObjectNotFoundException, failed to test a widget");
        }
    }

    private GuiWidget getNextWidget(GuiState currentScreen) {
        if (currentScreen.isFinished())
            return null;

        for (GuiWidget widget : currentScreen.getExecutableWidgets()) {
            if (!widget.getUiObject().exists() ||
                    (widget.isNAF() && !Config.sTestAlsoNAF)) {
                widget.setFinished(true); // If UI has changed or isNotHumanAccessible
                continue;
            }
            if (!widget.isFinished())
                return widget;
        }

        return null;
    }

    private boolean isNewTargetPkgScreen() {
        GuiState currentScreen = new GuiState();
        if (0 != currentScreen.getPkg().compareToIgnoreCase(Config.sTargetPackage))
            return false;
        return isNewScreen(currentScreen);
    }

    private boolean isNewScreen(GuiState currentScreen) {
        for (GuiState screen : sGuiGraph.getVertexValues()) {
            if (screen.equals(currentScreen)) {
                return false;
            }
        }
        return true;
    }

    // NOTE: useful after an action
    private boolean isNewOrUnfinishedScreen() {
        GuiState currentScreen = new GuiState();
        return isNewScreen(currentScreen) || !sGuiGraph.retrieve(currentScreen).isFinished();

    }

    private boolean isNewScreen() {
        GuiState currentScreen = new GuiState();
        return isNewScreen(currentScreen);
    }

    private boolean isAllScreensFinished() {
        for (GuiState screen : sGuiGraph.getVertexValues()) {
            if (!screen.isFinished()) {
                return false;
            }
        }
        return true;
    }

    private void logAllScreenInfo() {
        for (GuiState screen : sGuiGraph.getVertexValues()) {
            Log.d(TAG, "Screen[" + (screen.getId()) + "] " + screen.toString());
        }
        Log.d(TAG_DEBUG, "Root Screen id: " + sRootScreen.getId());
    }
}

