package com.vincenzojuniorforte.poliutodroid.core;

import android.app.Instrumentation;
import android.graphics.Rect;
import android.os.Build;
import android.support.test.uiautomator.UiDevice;
import android.support.test.uiautomator.UiObject;
import android.support.test.uiautomator.UiObjectNotFoundException;
import android.support.test.uiautomator.UiSelector;
import android.util.Log;
import android.view.accessibility.AccessibilityNodeInfo;
import android.view.accessibility.AccessibilityWindowInfo;

import com.vincenzojuniorforte.poliutodroid.Config;
import com.vincenzojuniorforte.poliutodroid.model.GuiWidget;
import com.vincenzojuniorforte.poliutodroid.model.structure.GenericTree;

import java.io.IOException;
import java.util.ArrayList;

/**
 * GuiHierarchyDumper class
 * <p>
 * It exploits UiAutomation in order to dump the layout hierarchy currently present on the device.
 */
public class GuiHierarchyDumper {

    //region Constant and Static Variables
    private static final int API_LEVEL_ACTUAL = Build.VERSION.SDK_INT + ("REL".equals(Build.VERSION.CODENAME) ? 0 : 1);
    private static final String LOG_TAG = GuiHierarchyDumper.class.getSimpleName();
    private static final String[] NAF_EXCLUDED_CLASSES = new String[]{
            android.widget.GridView.class.getName(), android.widget.GridLayout.class.getName(),
            android.widget.ListView.class.getName(), android.widget.TableLayout.class.getName()
    };
    //endregion

    //region Variables
    // Instrumentation object
    private Instrumentation mInstrumentation;
    //endregion

    //region Constructors

    /**
     * GuiHierarchyDumper constructor
     *
     * @param mInstrumentation Instrumentation object
     */
    public GuiHierarchyDumper(Instrumentation mInstrumentation) {
        this.mInstrumentation = mInstrumentation;
    }
    //endregion

    //region Public Methods

    /**
     * Starting from a root, perform a dump of the layout hierarchy currently present on the device
     *
     * @param device        device on which the dump operation is performed
     * @param targetPackage package to which the dumped layout belongs to
     * @return a generic tree composed by GuiWidget objects as nodes
     * @throws IOException
     */
    public GenericTree<GuiWidget> dump(UiDevice device, String targetPackage) throws IOException {
        GenericTree<GuiWidget> widgetTree = null;

        for (AccessibilityNodeInfo root : getWindowRoots(device)) {
            if (root.getPackageName().equals(targetPackage)) {
                // Retrieve UiObject Root
//                UiObject rootWidget = device.findObject(new UiSelector().instance(0));
                UiObject rootWidget = device.findObject(new UiSelector().packageName(targetPackage).instance(0));   // TODO: to be verified

                // Build Widget Root
                GuiWidget rootGuiWidget = new GuiWidget(root, rootWidget);
                // Set Tree Root CIS
                rootGuiWidget.setCIS(0);

                // Build Widget Tree
                GenericTree<GuiWidget> currentWidgetTree = new GenericTree<>();
                // Build Root Tree Node
                GenericTree<GuiWidget>.Node rootNode = currentWidgetTree.new Node(rootGuiWidget);
                // Set Tree Root
                currentWidgetTree.setRoot(rootNode);

                dumpNodeRec(root, 0, currentWidgetTree, null);

                // WORKAROUND: sometimes (e.g. when click on an filled edit box) happens that there are more then one AccessibilityNodeInfo with the same package; we take the one with the biggest number of child
                if (widgetTree == null || widgetTree.getNumberOfNodes() < currentWidgetTree.getNumberOfNodes()) {
                    widgetTree = currentWidgetTree;
                }
            }
        }

        return widgetTree;
    }
    //region Private Methods

    /**
     * Returns a list containing the root {@link AccessibilityNodeInfo}s for each active window
     */
    private AccessibilityNodeInfo[] getWindowRoots(UiDevice device) {
        device.waitForIdle();

        ArrayList<AccessibilityNodeInfo> ret = new ArrayList<>();
        // Support multi-window searches for API level 21 and up
        if (Build.VERSION.SDK_INT /* API_LEVEL_ACTUAL */ >= Build.VERSION_CODES.LOLLIPOP) {
            for (AccessibilityWindowInfo window : mInstrumentation.getUiAutomation().getWindows()) {
                AccessibilityNodeInfo root = window.getRoot();
                if (root == null) {
                    Log.w(LOG_TAG, String.format("Skipping null root node for window: %s",
                            window.toString()));
                    continue;
                }
                ret.add(root);
            }
            // Prior to API level 21 we can only access the active window
        } else {
            ret.add(mInstrumentation.getUiAutomation().getRootInActiveWindow());
        }
        return ret.toArray(new AccessibilityNodeInfo[ret.size()]);
    }

    /**
     * Starting from a given node, perform recursively a dump of the layout hierarchy currently present on the device
     *
     * @param node       node from which starting to recursively dump the layout hierarchy
     * @param index      auxiliary variable useful to uniquely identify a widget within a GuiState as CIS
     * @param widgetTree generic tree composed by GuiWidget objects as nodes built on-the-fly
     * @param parent     parent node useful for recursion
     * @throws IOException
     */
    private void dumpNodeRec(AccessibilityNodeInfo node, int index, GenericTree<GuiWidget> widgetTree, GenericTree<GuiWidget>.Node parent) throws IOException {

        UiObject currentUiObject = null;
        try {
            UiObject rootObject =
                    widgetTree.getRoot()    // get tree's root (Node object)
                            .getData()      // retrieve related GuiWidget object
                            .getUiObject(); // retrieve related getUiObject

            currentUiObject = parent == null ? rootObject : rootObject.getChild(new UiSelector().instance(widgetTree.getNumberOfNodes() - 1 /* Index numbering */));
        } catch (UiObjectNotFoundException | NullPointerException e) {
            e.printStackTrace();
            return;
        }

        GenericTree<GuiWidget>.Node currentTreeNode = null;

        if (parent != null) {
            // Build current GuiWidget object
            GuiWidget guiWidget = new GuiWidget(node, currentUiObject);
            guiWidget.setCIS(parent.getData().getCIS(), index);

            // Check for NAF
            if (!nafExcludedClass(node) && !nafCheck(node) && !humanVisible(guiWidget)) {
                guiWidget.setNAF(true);
                Log.i(LOG_TAG, String.format("Skipping invisible child: %s", node.toString()));
            }

            // Build current WidgetTreeNode object
            currentTreeNode = widgetTree.new Node(guiWidget);
            parent.addChild(currentTreeNode);
        } else {
            // The root node is already built and added to the widget tree
            currentTreeNode = widgetTree.getRoot();
            // Check for NAF
            if (!nafExcludedClass(node) && !nafCheck(node)) {
                currentTreeNode.getData().setNAF(true);
                Log.i(LOG_TAG, String.format("Skipping invisible child: %s", node.toString()));
            }
        }

        int count = node.getChildCount();
        for (int i = 0; i < count; i++) {
            AccessibilityNodeInfo child = node.getChild(i);
            if (child != null) {
                if (child.isVisibleToUser()) {
                    dumpNodeRec(child, i, widgetTree, currentTreeNode);
                    child.recycle();
                } else {
                    Log.i(LOG_TAG, String.format("Skipping invisible child: %s", child.toString()));
                }
            } else {
                Log.i(LOG_TAG, String.format("Null child %d/%d, parent: %s",
                        i, count, node.toString()));
            }
        }
    }

    /**
     * Verify if a GuiWidget is humanly visible (i.e., big enough)
     *
     * @param guiWidget GuiWidget to be verified
     * @return true if the GuiWidget is humanly visible; false otherwise
     */
    private boolean humanVisible(GuiWidget guiWidget) {
        try {
            Rect rect = guiWidget.getUiObject().getVisibleBounds();
            if ((rect.width() * rect.height()) / (float) Config.sDisplayResolution >= Config.sMinWidgetSurfacePercentage)
                return true;
        } catch (UiObjectNotFoundException e) {
            e.printStackTrace();
        }

        return false;
    }

    /**
     * The list of clSkipping null root node for windows to exclude my not be complete. We're attempting to
     * only reduce noise from standard layout classes that may be falsely
     * configured to accept clicks and are also enabled.
     *
     * @param node
     * @return true if node is excluded.
     */
    private boolean nafExcludedClass(AccessibilityNodeInfo node) {
        String className = safeCharSeqToString(node.getClassName());
        for (String excludedClassName : NAF_EXCLUDED_CLASSES) {
            if (className.endsWith(excludedClassName))
                return true;
        }
        return false;
    }

    /**
     * We're looking for UI controls that are enabled, clickable but have no
     * text nor content-description. Such controls configuration indicate an
     * interactive control is present in the UI and is most likely not
     * accessibility friendly. We refer to such controls here as NAF controls
     * (Not Accessibility Friendly)
     *
     * @param node
     * @return false if a node fails the check, true if all is OK
     */
    private boolean nafCheck(AccessibilityNodeInfo node) {
        boolean isNaf = /*node.isClickable() &&*/ node.isEnabled() && node.isVisibleToUser()
                && safeCharSeqToString(node.getContentDescription()).isEmpty()
                && safeCharSeqToString(node.getText()).isEmpty();

        if (!isNaf)
            return true;

        // check children since sometimes the containing element is clickable
        // and NAF but a child's text or description is available. Will assume
        // such layout as fine.
        return childNafCheck(node);
    }

    /**
     * This should be used when it's already determined that the node is NAF and
     * a further check of its children is in order. A node maybe a container
     * such as LinerLayout and may be set to be clickable but have no text or
     * content description but it is counting on one of its children to fulfill
     * the requirement for being accessibility friendly by having one or more of
     * its children fill the text or content-description. Such a combination is
     * considered by this dumper as acceptable for accessibility.
     *
     * @param node
     * @return false if node fails the check.
     */
    private boolean childNafCheck(AccessibilityNodeInfo node) {
        int childCount = node.getChildCount();
        for (int x = 0; x < childCount; x++) {
            AccessibilityNodeInfo childNode = node.getChild(x);

            if (!safeCharSeqToString(childNode.getContentDescription()).isEmpty()
                    || !safeCharSeqToString(childNode.getText()).isEmpty())
                return true;

            if (childNafCheck(childNode))
                return true;
        }
        return false;
    }

    private String safeCharSeqToString(CharSequence cs) {
        if (cs == null)
            return "";
        else {
            return stripInvalidXMLChars(cs);
        }
    }

    private String stripInvalidXMLChars(CharSequence cs) {
        StringBuffer ret = new StringBuffer();
        char ch;
        /* http://www.w3.org/TR/xml11/#charsets
        [#x1-#x8], [#xB-#xC], [#xE-#x1F], [#x7F-#x84], [#x86-#x9F], [#xFDD0-#xFDDF],
        [#x1FFFE-#x1FFFF], [#x2FFFE-#x2FFFF], [#x3FFFE-#x3FFFF],
        [#x4FFFE-#x4FFFF], [#x5FFFE-#x5FFFF], [#x6FFFE-#x6FFFF],
        [#x7FFFE-#x7FFFF], [#x8FFFE-#x8FFFF], [#x9FFFE-#x9FFFF],
        [#xAFFFE-#xAFFFF], [#xBFFFE-#xBFFFF], [#xCFFFE-#xCFFFF],
        [#xDFFFE-#xDFFFF], [#xEFFFE-#xEFFFF], [#xFFFFE-#xFFFFF],
        [#x10FFFE-#x10FFFF].
         */
        for (int i = 0; i < cs.length(); i++) {
            ch = cs.charAt(i);

            if ((ch >= 0x1 && ch <= 0x8) || (ch >= 0xB && ch <= 0xC) || (ch >= 0xE && ch <= 0x1F) ||
                    (ch >= 0x7F && ch <= 0x84) || (ch >= 0x86 && ch <= 0x9f) ||
                    (ch >= 0xFDD0 && ch <= 0xFDDF) || (ch >= 0x1FFFE && ch <= 0x1FFFF) ||
                    (ch >= 0x2FFFE && ch <= 0x2FFFF) || (ch >= 0x3FFFE && ch <= 0x3FFFF) ||
                    (ch >= 0x4FFFE && ch <= 0x4FFFF) || (ch >= 0x5FFFE && ch <= 0x5FFFF) ||
                    (ch >= 0x6FFFE && ch <= 0x6FFFF) || (ch >= 0x7FFFE && ch <= 0x7FFFF) ||
                    (ch >= 0x8FFFE && ch <= 0x8FFFF) || (ch >= 0x9FFFE && ch <= 0x9FFFF) ||
                    (ch >= 0xAFFFE && ch <= 0xAFFFF) || (ch >= 0xBFFFE && ch <= 0xBFFFF) ||
                    (ch >= 0xCFFFE && ch <= 0xCFFFF) || (ch >= 0xDFFFE && ch <= 0xDFFFF) ||
                    (ch >= 0xEFFFE && ch <= 0xEFFFF) || (ch >= 0xFFFFE && ch <= 0xFFFFF) ||
                    (ch >= 0x10FFFE && ch <= 0x10FFFF))
                ret.append(".");
            else
                ret.append(ch);
        }
        return ret.toString();
    }
    //endregion
}